import optparse
import os
import shutil
import subprocess
import sys
import tempfile

from django.conf import settings
from django.utils import unittest
from django.test import runner as dj_runner
from gargoyle import gargoyle
import subunit
import testtools
from ucitests import (
    filters,
    results,
    runners,
    suites,
)

try:
    import cProfile as profile
except ImportError:
    import profile


class SSOTestSuite(suites.TestSuite):
    """Test suite for the canonical identity provider project.

    This centralizes the setup for the test environment and applies to all
    tests. Use only for setups that can't be done at a lower level.

    """

    def setUp(self):
        super(SSOTestSuite, self).setUp()
        # gargoyle caches settings. Setting timeout to 0 prevents this,
        # increasing test isolation.
        gargoyle.timeout = 0
        # We need unique identifiers for several pieces, the process pid is a
        # good source of uniqueness which works for concurrent runs on the same
        # host as well as subprocesses for a single run.
        pid = os.getpid()
        # We want the db to be unique for each worker (or the master if no
        # concurrency is involved).
        for k, v in settings.DATABASES.items():
            v['TEST']['NAME'] = 'test_{}_{}'.format(v['NAME'], pid)
        # Caches need to be unique, we get there by using a unique key prefix
        settings.CACHES['default']['KEY_PREFIX'] = '{}-'.format(pid)
        # Make oops dir unique
        settings.OOPSES['publishers'][0]['error_dir'] = os.path.join(
            settings.LOGS_DIR, 'www-oops-{}'.format(pid))
        # Read-only mode relies on the 'db.readonly' file in
        # DBFAILOVER_FLAG_DIR, this needs to be unique
        settings.DBFAILOVER_FLAG_DIR = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, settings.DBFAILOVER_FLAG_DIR, True)
        # 'interactive' in the call below is used internally to autoclobber dbs
        # (using 'not interactive'). 'verbosity' is set to 1 to only display
        # creating/deletion of the databases.
        old_config = dj_runner.setup_databases(verbosity=1,
                                               interactive=False)
        self.addCleanup(self.teardown_databases, old_config)

    def teardown_databases(self, old_config):
        old_names, mirrors = old_config
        for connection, old_name, destroy in old_names:
            if destroy:
                connection.creation.destroy_test_db(old_name, verbosity=1)


class TestRunner(dj_runner.DiscoverRunner):

    option_list = dj_runner.DiscoverRunner.option_list + (
        optparse.make_option(
            '-c', '--concurrency',
            action='store', dest='concurrency', type='int',
            default=1,
            help='How many concurrent processes to run the tests.'),
        optparse.make_option(
            '-l', '--list',
            action='store_true', dest='list_only',
            default=False,
            help='List test ids without running them.'),
        optparse.make_option(
            '-i', '--include',
            action='store', dest='include',
            default=None,
            help='Tests matching this regexp will be run.'),
        optparse.make_option(
            '-x', '--exclude',
            action='store', dest='exclude',
            default=None,
            help='Tests matching this regexp will not be run.'),
        optparse.make_option(
            '--load',
            action='store', dest='load_file',
            default=None,
            help='A file name containing test ids to be run.'),
        optparse.make_option(
            '-f', '--format',
            action='store', dest='format', type='choice',
            choices=['text', 'subunit'], default='text',
            help='Output format for the test results.'),
        optparse.make_option(
            '-s', '--sub-runner',
            action='store', dest='sub_runner', type='choice',
            choices=['fork', 'lxc'], default='fork',
            help='The concurrent test runner to use.'),
    )

    def __init__(self, pattern=None, top_level=None, verbosity=1,
                 interactive=True, failfast=False, concurrency=1,
                 list_only=False, include=None, exclude=None,
                 load_file=None, format=None,
                 sub_runner=None,
                 output=None,
                 **kwargs):
        super(TestRunner, self).__init__(
            pattern, top_level, verbosity=verbosity, interactive=interactive,
            failfast=failfast, **kwargs)
        self.list_only = list_only
        if output is None:
            output = sys.stdout
        self.output = output
        if include is not None:
            self.includes = include.split(':')
        else:
            self.includes = []
        if exclude is not None:
            self.excludes = exclude.split(':')
        else:
            self.excludes = []
        self.concurrency = concurrency
        self.load_file = load_file
        self.format = format
        self.sub_runner = sub_runner

    def setup_databases(self, **kwargs):
        """Override base class to defer the db setup.

        The db setup id done in SSOTestSuite so each worker can use its own
        db.

        """
        # Return an empty but valid value so teardown_databases does nothing
        # when invoked from the base class run_suite().
        return ([], [])

    def build_suite(self, test_labels=None, extra_tests=None, **kwargs):
        if not test_labels:
            # Without this, only the tests in the acceptance folder
            # will be tried to be run when no labels are specified.
            test_labels = ('src',)
        suite = super(TestRunner, self).build_suite(
            test_labels, extra_tests, **kwargs)
        suite = filters.include_regexps(self.includes, suite)
        suite = filters.exclude_regexps(self.excludes, suite)
        if self.load_file is not None:
            # Load the file and filter the suite
            with open(self.load_file, 'r') as f:
                id_list = filters.TestIdList(filters.load_test_id_list(f))
                suite = filters.include_id_list(id_list, suite)
        wrapping_suite = SSOTestSuite(suite)
        return wrapping_suite

    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        # Override base class so we can list the tests without the costly
        # setup.
        if self.list_only:
            suite = self.build_suite(test_labels, extra_tests)
            for t in testtools.testsuite.iterate_tests(suite):
                self.output.write(t.id() + '\n')
            return 0
        # Otherwise let the base class handle the setup and the run
        return super(TestRunner, self).run_tests(test_labels, extra_tests=None,
                                                 **kwargs)

    def run_suite(self, suite, **kwargs):
        initial_test_nb = suite.countTestCases()
        if self.concurrency > 1:
            if self.sub_runner == 'fork':
                concurrent_runner = runners.run_suite_forked
            elif self.sub_runner == 'lxc':
                concurrent_runner = run_suite_in_container
            concurrent_suite = testtools.ConcurrentTestSuite(
                suite, runners.split_suite_for(self.concurrency,
                                               concurrent_runner))
        else:
            concurrent_suite = suite
        if self.format == 'text':
            result = results.TextResult(sys.stdout,
                                        verbosity=self.verbosity,
                                        failfast=self.failfast)
        else:
            result = subunit.TestProtocolClient(sys.stdout)
        result.startTestRun()
        try:
            concurrent_suite.run(result)
            self.check_all_tests_executed(initial_test_nb, result)
        except subprocess.CalledProcessError as e:
            sys.stderr.write('{}'.format(e))
            raise
        finally:
            result.stopTestRun()
        return result

    def check_all_tests_executed(self, initial_test_nb, result):
        # In very rare failure modes, tests that should be run may not
        # increment 'testsRun'. This sanity check will catch them.
        try:
            if initial_test_nb != result.testsRun:
                raise AssertionError(
                    'Expecting {}, got {} tests run'.format(
                        initial_test_nb, result.testsRun))
        except:
            # Defined here to get a revealing class path without risking test
            # discovery to pick us.
            class RunSuiteIntegrity(unittest.TestCase):
                # 'run' is just an existing method, we won't execute that test
                runTest = unittest.TestCase.run

            t = RunSuiteIntegrity('run')
            result.errors.append((t, result._exc_info_to_string(sys.exc_info(),
                                                                t)))
            # The above avoid a spurious 'ERROR' on the output stream if we use
            # the following (clearer and not relying on private details) call
            # result.addError(RunSuiteIntegrity('run'), sys.exc_info())


# FIXME: More of the following should be upstreamed to uci-{tests,vms}. Part of
# it has been stolen from there to start with. -- vila 2015-06-09


def subprocess_run(args):
    try:
        return subprocess.check_output(args,
                                       stderr=subprocess.STDOUT,
                                       stdin=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        # By default, the bare exception won't reveal the output from the
        # failed command, we do before re-raising.
        sys.stderr.write('Command {} failed with {}'.format(
            ' '.join(args), e.output))
        raise


class TestInContainer(object):

    def __init__(self, worker_name, test_list_path, worker_num):
        super(TestInContainer, self).__init__()
        self.worker_name = worker_name
        self.test_list_path = test_list_path
        self.worker_num = worker_num

    def __call__(self, result=None):
        # comply with unittests.TestCase API
        return self.run(result)

    def run(self, result):
        here = os.getcwd()
        # Spawn the ephemeral container
        subprocess_run(['uci-vms', 'start', self.worker_name])
        try:
            # Finish container setup (food for uci-vms there)
            subprocess_run(['uci-vms', 'shell', self.worker_name,
                            '(', 'cd', here, '&&',
                            'make', 'start-db',
                            ')'])
            # Run the tests, connecting the pipes
            cmd = ['uci-vms', 'shell', self.worker_name,
                   # And now we run the tests inside the container
                   '(', 'cd', here, '&&',
                   'make', 'test',
                   'ARGS="-f subunit --load {}"'.format(self.test_list_path),
                   ')']
            process = subprocess.Popen(cmd,
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       # line buffering for subunit
                                       bufsize=1)
            # run the subunit test that collect the real tests results
            test = TestInSubprocess(process)
            test.run(result)
        finally:
            # Whatever happens, cleanup.
            # Note that for debugging isolation issues, it may help to start
            # with the 'self.test_list_path' file, so commenting the following
            # line will preserve all lists.
            os.remove(self.test_list_path)
            # And tear down the ephemeral container
            subprocess_run(['uci-vms', 'stop', self.worker_name])


class TestInSubprocess(subunit.ProtocolTestCase):

    def __init__(self, process):
        super(TestInSubprocess, self).__init__(process.stdout)
        self.process = process
        self.process.stdin.close()

    def run(self, result):
        try:
            super(TestInSubprocess, self).run(result)
        finally:
            self.process.wait()


def run_suite_in_container(unique, child_suite, parent_suite):
    """Execute a test suite into a container.

    :param unique: A number uniquely identifying the suite/container.

    :param child_suite: A TestSuite object to be run in the container.

    :param parent_suite: A TestSuite object where the parent test collecting
        the child results should be added.
    """
    worker_name = 'sso-worker-{:>02d}'.format(unique)
    # FIXME: We use a file in the current directory as a cheap way to share it
    # with the container which bind mount the user HOME. It would be better to
    # upload it to the container instead but it's more involved. -- vila
    # 2015-06-09
    test_list_path = '{}.test-list'.format(worker_name)
    with open(test_list_path, 'w') as tl:
        for t in filters.iter_flat(child_suite):
            tl.write('{}\n'.format(t.id()))
    test = TestInContainer(worker_name, test_list_path, unique)
    parent_suite.addTest(test)


class ProfilingTestSuiteRunner(TestRunner):

    def run_suite(self, suite, **kwargs):
        runner = unittest.TextTestRunner(
            verbosity=self.verbosity, failfast=self.failfast).run

        profile.runctx(
            'result = run_tests(suite)',
            {
                'run_tests': runner,
                'suite': suite,
            },
            locals(),
            getattr(settings, 'TEST_PROFILE', None),
        )
        return locals()['result']
