# Copyright 2010, 2012 Canonical Ltd.  This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from __future__ import unicode_literals

import logging
import os

from urllib import urlencode

from convoy.combo import combine_files, parse_qs
from django.conf import settings
from django.contrib import auth, messages
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.http import (
    Http404,
    HttpResponse,
    HttpResponseNotAllowed,
    HttpResponseRedirect,
    HttpResponseNotFound,
)
from django.shortcuts import (
    get_object_or_404,
    redirect,
    render_to_response,
)
from django.template import RequestContext
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.utils.http import urlquote
from django.utils.translation import ugettext_lazy as _
from django.views.generic.base import View, TemplateResponseMixin
from django.views.decorators.cache import cache_control
from django.views.decorators.gzip import gzip_page
from gargoyle import gargoyle

from identityprovider import emailutils
from identityprovider.forms import (
    LoginForm,
    NewAccountForm,
    ResetPasswordForm,
    TokenForm,
    TwoFactorForm,
    TwoFactorLoginForm,
)
from identityprovider.login import (
    AuthenticationError,
    PasswordPolicyError,
    authenticate_user,
)
from identityprovider.models import (
    Account,
    AuthToken,
    EmailAddress,
    OpenIDRPConfig,
    get_type_of_token,
    verify_token_string,
)
from identityprovider.models import twofactor
from identityprovider.models.captcha import Captcha, VerifyCaptchaError
from identityprovider.models.const import AccountStatus, AuthTokenType
from identityprovider.signals import login_failed, login_succeeded
from identityprovider.utils import (
    polite_form_errors,
    add_non_field_error,
)
from identityprovider.stats import stats
from identityprovider.timeline_helpers import get_request_timing_function
from identityprovider.validators import PASSWORD_LEAKED
from identityprovider.views.utils import (
    get_rpconfig_from_request,
    is_safe_redirect_url,
)

from webui.decorators import (
    check_readonly,
    dont_cache,
    redirect_home_if_logged_in,
    limitlogin,
    ratelimit_extra,
    requires_cookies,
    require_twofactor_enabled,
)
from webui.views.utils import add_captcha_settings
from webui.views import registration


ACCOUNT_CREATED = _("Your account was created successfully")
TWOFACTOR_REQUIRED = _(
    'Your account requires 2-factor authentication but you have no '
    'authentication devices in the system. ')

logger = logging.getLogger(__name__)


# Note: using TemplateResponseMixin rather than TemplateView, as it doesn't
# quite match our use case here
class LoginBaseView(TemplateResponseMixin, View):
    """Base class for login views.

    Handles context creation, rendering, initial set up, and error handling."""

    def setup(self, request, token, rpconfig):
        if not rpconfig:
            rpconfig = get_rpconfig_from_request(request, token)
        return rpconfig

    def get_context(self, request, token, rpconfig, form):
        context = {
            'form': form,
            'next': request.GET.get('next'),
            'rpconfig': rpconfig,
            'token': token,
        }
        return context

    def render(self, request, **kwargs):
        context = self.get_context(request, **kwargs)
        context = RequestContext(request, context)

        return self.render_to_response(context)

    def display_errors(self, request, rpconfig, form, error=None, **kwargs):
        """Track and display nice errors, including extra ones."""
        # track login form errors
        stats.increment('flows.login', key='error', rpconfig=rpconfig)
        polite_form_errors(form._errors)
        if error:
            add_non_field_error(form, error)
        return self.render(request, rpconfig=rpconfig, form=form, **kwargs)


class LoginView(LoginBaseView):

    template_name = 'registration/login.html'

    def get_context(
            self, request, token, rpconfig, form, create_account_form=None):
        context = super(LoginView, self).get_context(
            request, token=token, rpconfig=rpconfig, form=form)
        # add captcha and account creation form
        context['captcha_required'] = gargoyle.is_active('CAPTCHA', request)
        context = add_captcha_settings(context)
        context['create_account_form'] = create_account_form
        return context

    def get_login_type(self, request, token, rpconfig):
        required = twofactor.site_requires_twofactor_auth(request, token,
                                                          rpconfig)
        return required, TwoFactorLoginForm if required else LoginForm

    def get(self, request, token=None, rpconfig=None):
        rpconfig = self.setup(request, token, rpconfig)
        required2f, form_cls = self.get_login_type(request, token, rpconfig)
        # track login attempts
        stats.increment('flows.login', key='requested', rpconfig=rpconfig)

        create_account_form = NewAccountForm()

        return self.render(
            request, token=token, rpconfig=rpconfig, form=form_cls(),
            create_account_form=create_account_form)

    def post(self, request, token=None, rpconfig=None):
        rpconfig = self.setup(request, token, rpconfig)
        result = self.get_login_type(request, token, rpconfig)
        site_twofactor, form_cls = result

        form = form_cls(request.POST)
        create_account_form = NewAccountForm()
        if not form.is_valid():
            return self.display_errors(
                request, token=token, rpconfig=rpconfig, form=form,
                create_account_form=create_account_form)

        # attempt user login
        email = form.cleaned_data['email']
        password = form.cleaned_data['password']
        try:
            account = authenticate_user(
                email, password, rpconfig=rpconfig, request=request)
        except PasswordPolicyError as e:
            auth_token = AuthToken.objects.create(
                requester=e.account, email=email,
                requester_email=email,
                token_type=AuthTokenType.PASSWORDRECOVERY)
            auth_token.redirection_url = request.META.get('HTTP_REFERER')
            auth_token.save()
            token = auth_token.raw_token
            reset_password_url = reverse(
                'reset_password',
                kwargs={'authtoken': token, 'email_address': email})

            message = unicode(e)
            if (e.account.need_password_reset or
                    unicode(PASSWORD_LEAKED) in message):
                error_msg = _("You need to change your password")
                context = RequestContext(request)
                message = render_to_string(
                    'common/password_leaked_warning.txt', context)
            else:
                error_msg = _(
                    "Your password doesn't meet security constraints")

            messages.warning(request, error_msg)
            messages.info(request, message)
            return HttpResponseRedirect(reset_password_url)
        except AuthenticationError as e:
            # track failed login attempt
            login_failed.send(sender=self, request=request,
                              credentials=dict(email=email, password=password))
            return self.display_errors(
                request, token=token, rpconfig=rpconfig, form=form, error=e,
                create_account_form=create_account_form)
        else:
            # Log a webui authentication success (web ui asked for
            # login/password)
            login_succeeded.send(
                sender=self, user=account, request=request, email=email)
            auth.login(request, account)

        next_url = request.POST.get('next')

        # handle case where the user's account requires two factor but we
        # didn't know that until we authenticated them
        oath_token = form.cleaned_data.get('oath_token')
        if (not oath_token and
                twofactor.user_requires_twofactor_auth(request, account)):
            kwargs = {'token': token} if token else {}
            url = reverse('twofactor', kwargs=kwargs)
            if next_url and is_safe_redirect_url(next_url):
                # preserve next_url across get
                url += "?next=" + urlquote(next_url)
            return HttpResponseRedirect(url)

        if site_twofactor or oath_token:
            try:
                twofactor.authenticate(account, oath_token)
                twofactor.login(request)
            except AuthenticationError as e:
                auth.logout(request)
                return self.display_errors(
                    request, token=token, rpconfig=rpconfig, form=form,
                    error=e, create_account_form=create_account_form)

        # track successful logins
        stats.increment('flows.login', key='success', rpconfig=rpconfig)

        # only clear limits if both succesfully auth
        limitlogin().reset_count(request)

        if next_url and is_safe_redirect_url(next_url):
            url = next_url
        elif token:
            url = '/%s/' % token
        else:
            url = '/'
        return HttpResponseRedirect(url)

    # To decorate class view methods with decorators designed for regular views
    # we must overwrite and decorate the dispatch method
    # OMFG. Srsly? :(
    # Docs: https://docs.djangoproject.com/en/dev/topics/class-based-views/
    @method_decorator(redirect_home_if_logged_in)
    @method_decorator(dont_cache)
    @method_decorator(limitlogin())
    @method_decorator(requires_cookies)
    def dispatch(self, *args, **kwargs):
        return super(LoginView, self).dispatch(*args, **kwargs)


login_view = LoginView.as_view()


class TwoFactorView(LoginBaseView):
    template_name = 'registration/twofactor.html'

    def get(self, request, token=None, rpconfig=None):
        if not request.user.has_twofactor_devices():
            messages.error(request, TWOFACTOR_REQUIRED)
            # ensure the template does not have the form
            form = None
        else:
            form = TwoFactorForm()
        return self.render(request, token=token, rpconfig=rpconfig, form=form)

    def display_errors(self, account, request, **kwargs):
        notif_threshold = settings.TWOFACTOR_FAILURE_NOTIFICATION_THRESHOLD
        filter_teams = settings.TWOFACTOR_FAILURE_NOTIFICATION_TEAMS
        if (account.twofactor_attempts >= notif_threshold and
                account.person_in_any_team(filter_teams)):
            emailutils.send_twofactor_failure_warning(account, request.META)

        return super(TwoFactorView, self).display_errors(request, **kwargs)

    def _suspend_too_many_2fa_failures(self, account, request):
        logger.warning(
            'Suspending account %r, %r due to too many twofactor '
            'failures', account.openid_identifier, account.id)
        account.suspend()
        auth.logout(request)
        if account.person_in_any_team(
                settings.TWOFACTOR_FAILURE_NOTIFICATION_TEAMS):
            reason = 'Too many 2FA failures (%s)' % account.twofactor_attempts
            emailutils.send_account_suspended_warning(account, reason)
        return suspended(request)

    def post(self, request, token=None, rpconfig=None):
        account = Account.objects.select_for_update().get(
            id=request.user.id)
        request.user = account
        account.increment_twofactor_attempts()

        form = TwoFactorForm(request.POST)
        if not form.is_valid():
            return self.display_errors(
                account, request, token=token, rpconfig=rpconfig, form=form)

        try:
            twofactor.authenticate(account, form.cleaned_data['oath_token'])
            twofactor.login(request)
            stats.increment('flows.login', key='success', rpconfig=rpconfig)
        except AuthenticationError as e:
            if account.twofactor_attempts >= settings.TWOFACTOR_MAX_ATTEMPTS:
                response = self._suspend_too_many_2fa_failures(
                    account, request)
            else:
                response = self.display_errors(
                    account, request, token=token, rpconfig=rpconfig,
                    form=form, error=e)
            return response
        else:
            account.reset_twofactor_attempts()

        limitlogin().reset_count(request)

        next_url = request.POST.get('next')
        if next_url and is_safe_redirect_url(next_url):
            return HttpResponseRedirect(next_url)
        elif token:
            return HttpResponseRedirect('/%s/' % token)
        return HttpResponseRedirect('/')

    @method_decorator(dont_cache)
    @method_decorator(limitlogin())
    # for this page, we need to be logged in but NOT twofactored
    @method_decorator(login_required)
    @method_decorator(require_twofactor_enabled)
    def dispatch(self, *args, **kwargs):
        if 'rpconfig' not in kwargs:
            request = args[0]
            token = kwargs.get('token')
            kwargs['rpconfig'] = get_rpconfig_from_request(request, token)
        return super(TwoFactorView, self).dispatch(*args, **kwargs)


twofactor_view = TwoFactorView.as_view()


class LogoutView(TemplateResponseMixin, View):

    template_name = 'registration/logout.html'

    def get_return_to_url_and_site_name(self, request):
        return_to = request.REQUEST.get('return_to')
        # This rpconfig may not be the original from the request, but that's OK
        rpconfig = OpenIDRPConfig.objects.for_url(return_to)
        # Due to security reasons, we will not allow return_to URLs that are
        # not within an RPConfig's trust_root. To check this, always ensure
        # the trust_root ends with /
        result = (None, None)
        if rpconfig:
            trust_root = rpconfig.trust_root.rstrip('/') + '/'
            if return_to.startswith(trust_root):
                result = return_to, rpconfig.displayname or return_to
        return result

    def get_other_sites(self, user, trust_root):
        get_active_sessions = getattr(user, 'sites_with_active_sessions', None)
        if get_active_sessions is not None:
            result = [
                site for site in get_active_sessions()
                if trust_root != site.trust_root]
        else:
            result = []
        return result

    def get(self, request, *args, **kwargs):
        return self.post(request, *args, **kwargs)

    def post(self, request, token=None):
        return_to, site_name = self.get_return_to_url_and_site_name(request)

        # get language before logging the user out
        language = getattr(request.user, 'preferredlanguage', None)
        # requires user to be still logged in
        other_sites = self.get_other_sites(request.user, return_to)

        # We don't want to lose session[token] when we log the user out
        raw_orequest = request.session.get(token, None)
        auth.logout(request)
        twofactor.logout(request)

        # restore the token and the openid request
        if token is not None and raw_orequest is not None:
            request.session[token] = raw_orequest

        # if the requester wants to immediatly redirect the user, let's do it.
        # We already ensured return_to is a safe url to navigate.
        if return_to and request.REQUEST.get('return_now', False):
            response = HttpResponseRedirect(return_to)
        else:
            context = {
                'return_to_url': return_to,
                'return_to_site_name': site_name,
                'other_sites': other_sites,
                # This rpconfig is the one originating the logout request
                'rpconfig': get_rpconfig_from_request(request, token),
            }
            response = self.render_to_response(context)

        # reset language after session was cleared out on logout
        if language:
            response.set_cookie(settings.LANGUAGE_COOKIE_NAME, language)
        return response


logout = LogoutView.as_view()


def enter_token(request, token=None):
    confirmation_code = request.REQUEST.get('confirmation_code')
    email = request.REQUEST.get('email') or request.session.get('token_email')
    params = {'confirmation_code': confirmation_code, 'email': email}
    if request.method == 'GET':
        form = TokenForm(initial=params)
    elif request.method == 'POST':
        form = TokenForm(params)
        if form.is_valid():
            confirmation = form.cleaned_data['confirmation']
            return _handle_confirmation(
                confirmation_type=confirmation.token_type,
                confirmation_code=confirmation_code, email=email, token=token)
    else:
        return HttpResponseNotAllowed(['GET', 'POST'])
    context = RequestContext(request, {
        'form': form,
        'token': token
    })
    return render_to_response('enter_token.html', context)


def _redirect_to_enter_token(confirmation_code=None, email=None):
    params = {}
    if confirmation_code is not None:
        params['confirmation_code'] = confirmation_code
    if email is not None:
        params['email'] = email
    return HttpResponseRedirect('/+enter_token?%s' % urlencode(params))


def claim_token(request, authtoken):
    email = request.GET.get('email') or request.session.get('token_email')
    if email is None:
        return _redirect_to_enter_token(confirmation_code=authtoken)
    token_type = get_type_of_token(authtoken)
    return _handle_confirmation(
        confirmation_type=token_type, confirmation_code=authtoken, email=email)


def _handle_confirmation(
        confirmation_type, confirmation_code, email, token=None):
    """Handle a confirmation for an action requested on 'email'.

    Either finish the work in this view, or redirect to a view that can. Also
    accepts an optional OpenID-transaction token.

    """

    args = {
        'authtoken': confirmation_code,
        'email_address': email
    }
    if token is not None:
        args['token'] = token

    if confirmation_type == AuthTokenType.PASSWORDRECOVERY:
        view = 'reset_password'
    elif confirmation_type == AuthTokenType.VALIDATEEMAIL:
        view = 'confirm_email'
    elif confirmation_type == AuthTokenType.INVALIDATEEMAIL:
        view = 'invalidate_email'
    else:
        msg = ('Unknown type %s for confirmation code "%s"' %
               (confirmation_type, confirmation_code))
        logger.error(msg)
        raise Http404(msg)

    url = reverse(view, kwargs=args)
    return HttpResponseRedirect(url)


@check_readonly
def new_account(request, token=None):
    return registration.new_account(request, token)


def _verify_captcha_response(template, request, form):
    captcha = Captcha(request.POST.get('recaptcha_challenge_field'))
    captcha_solution = request.POST.get('recaptcha_response_field')
    email = request.POST.get('email', '')
    ip_addr = request.environ["REMOTE_ADDR"]
    try:
        timer_fn = get_request_timing_function(request)
        verified = captcha.verify(captcha_solution, ip_addr, email,
                                  timer=timer_fn)
        if verified:
            return None
    except VerifyCaptchaError:
        logger.exception("reCaptcha connection error")

    # not verified
    return render_to_response(
        template,
        RequestContext(request, add_captcha_settings({
            'form': form,
            'captcha_error': ('&error=%s' % captcha.message),
            'captcha_required': True})))


def confirm_email(request, authtoken, email_address, token=None):
    if not twofactor.is_authenticated(request):
        messages.warning(request,
                         _('Please log in to use this confirmation code'))
        next_url = urlquote(request.get_full_path())
        return HttpResponseRedirect('%s?next=%s' % (settings.LOGIN_URL,
                                                    next_url))
    atrequest = verify_token_string(authtoken, email_address)
    if (atrequest is None or
            atrequest.token_type != AuthTokenType.VALIDATEEMAIL):
        return HttpResponseRedirect('/+bad-token')

    email = get_object_or_404(EmailAddress, email__iexact=atrequest.email)
    if request.user.id != email.account.id:
        # The user is authenticated to a different account.
        # Potentially, the token was leaked or intercepted.  Let's
        # delete it just in case.  The real user can generate another
        # token easily enough.
        atrequest.delete()
        # return a 404 response instead of raising so that
        # the TransactionMiddleware will not rollback the dirty transaction
        return HttpResponseNotFound()

    if request.method == 'POST':
        # only confirm the email address if the form was submitted
        email.validate()
        atrequest.consume()
        msg = _("The email address {email} has been validated")
        messages.success(request, msg.format(email=email))
        return HttpResponseRedirect(
            atrequest.redirection_url or reverse('account-index')
        )

    # form was not submitted
    context = RequestContext(request, {'email': email.email, 'token': token})
    return render_to_response('account/confirm_new_email.html', context)


def bad_token(request):
    return TemplateResponse(request, 'bad_token.html')


def logout_to_confirm(request):
    return TemplateResponse(request, 'registration/logout_to_confirm.html')


def deactivated(request):
    return TemplateResponse(request, 'account/deactivated.html')


def suspended(request):
    return TemplateResponse(request, 'account/suspended.html')


@ratelimit_extra(requests=20, minutes=60, key_field='email')
def reset_password(request, authtoken, email_address, token=None):
    # Need this for displaying metrics per RP
    rpconfig = get_rpconfig_from_request(request, token)

    atrequest = verify_token_string(authtoken, email_address)
    account = atrequest.requester if atrequest else None

    if (atrequest is None or
            atrequest.token_type != AuthTokenType.PASSWORDRECOVERY or
            not account.can_reset_password):
        # track bad token error
        stats.increment('flows.forgot_password', key='error.token',
                        rpconfig=rpconfig)
        # we hide the fact the the account is inactive, to avoid
        # exposing valid accounts. however, deactivated accounts can be
        # reactivated.
        return HttpResponseRedirect('/+bad-token')

    if request.method == 'POST':
        form = ResetPasswordForm(request.POST, account=account)
        if form.is_valid():
            password = form.cleaned_data['password']
            try:
                email_obj = EmailAddress.objects.get(
                    account=account, email__iexact=atrequest.requester_email)
            except EmailAddress.DoesNotExist:
                logger.warning(
                    'Could not reset password, invalid email address provided '
                    'in the token: %s (account %s)',
                    atrequest.requester_email, account.id)
                return HttpResponseRedirect('/+bad-token')
            # since the user must have received this token url via this email,
            # we know it's valid.
            email_obj.validate()
            if not account.is_active and account.can_reactivate:
                account.preferredemail = email_obj
                # reactivate the account
                account.status = AccountStatus.ACTIVE
                account.save()
                msg = _("Your account was reactivated, and the preferred "
                        "email address was set to {email}.")
                messages.info(request, msg.format(email=email_obj.email))

            email = account.preferredemail.email
            account.set_password(password)
            user = auth.authenticate(username=email, password=password,
                                     request=request)
            auth.login(request, user)
            atrequest.consume()
            # track password recovery success
            stats.increment('flows.forgot_password', key='success',
                            rpconfig=rpconfig)
            url = atrequest.redirection_url or reverse('account-index')
            return HttpResponseRedirect(url)

        # track form errors
        stats.increment('flows.forgot_password', key='error.form',
                        rpconfig=rpconfig)
    else:
        form = ResetPasswordForm(account=account)

    password_leaked = unicode(PASSWORD_LEAKED) in form.errors.get(
        'password', '')
    if password_leaked:
        context = RequestContext(request)
        error_msg = render_to_string('common/password_leaked_warning.txt',
                                     context)
        messages.warning(request, error_msg)

    context = RequestContext(
        request, {'form': form, 'token': token, 'email': email_address})
    return render_to_response('registration/reset_password.html', context)


def static_page(request, page_name, context=None):
    context = RequestContext(request, context)
    return render_to_response('static/%s.html' % page_name, context)


# combo view is only used for YUI, which will never change, so safe to cache
# this for a long time (1 year).
@cache_control(max_age=60 * 60 * 24 * 365)
@gzip_page
def combo_view(request):
    """Handle a request for combining a set of files."""
    fnames = parse_qs(request.META.get("QUERY_STRING", ""))
    content_type = "text/plain"

    if fnames:
        if fnames[0].endswith(".js"):
            content_type = "text/javascript"
        elif fnames[0].endswith(".css"):
            content_type = "text/css"

        content = combine_files(
            fnames, os.path.abspath(settings.STATIC_ROOT),
            resource_prefix=settings.STATIC_URL, rewrite_urls=True)
        # explicitly force no session usage, to avoid Vary: Cookie being added,
        # which breaks *all* caching
        request.session.accessed = request.session.modified = False
        return HttpResponse(
            content_type=content_type, status=200,
            content=b"".join(content))
    return HttpResponse(content_type=content_type, status=404)


def permanent_redirect_home(request, *args, **kwargs):
    return redirect('/', permanent=True)
